/*
 * tcpWinSink.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "tcpWinSink.h"
#include "tcpWin.h"

#include <unistd.h> // for close

/*
 * Variables from dependencies, i.e., other plugins, MUST be declared weak,
 * in order to prevent dlopen() from trying to resolve them. If the symbols
 * are missing, it means the required dependency was not loaded. The error
 * will be reported by loadPlugins.c when checking for the dependencies
 * listed in the get_dependencies() or T2_PLUGIN_INIT_WITH_DEPS() function.
 */

extern gwz_t *gwzP __attribute__((weak));  // define pointer to tcpWin window size threshold count structure


/*
 * Static variables are only visible in this file
 */

static int sfd;                   // socket handle
static struct sockaddr_in server; // struct for socket to connect to server
static struct timeval tmrk;       // time marker
static char sBuf[TWS_MAXSBUF];    // send buffer


/*
 * Static functions prototypes
 */


// Tranalyzer functions

/*
 * This describes the plugin name, version, major and minor version of
 * Tranalyzer required and dependencies
 */
T2_PLUGIN_INIT_WITH_DEPS("tcpWinSink", "0.9.0", 0, 9, "tcpWin");


/*
 * This function is called before processing any packet.
 */
void t2Init() {
    const struct hostent * const host = gethostbyname(TWS_SERVADD);
    if (UNLIKELY(!host)) {
        T2_PFATAL(plugin_name, "gethostbyname() failed for '%s'", TWS_SERVADD);
    }

    server.sin_addr = *(struct in_addr*)host->h_addr;
    server.sin_family = AF_INET;
    server.sin_port = htons(TWS_DPORT);

#if TWS_SOCKTYPE == 1 // TCP socket
    if (UNLIKELY(!(sfd = socket(AF_INET, SOCK_STREAM, 0)))) {
#else // TWS_SOCKTYPE == 0 // UDP socket
    if (UNLIKELY(!(sfd = socket(AF_INET, SOCK_DGRAM, 0)))) {
#endif // TWS_SOCKTYPE
        T2_PFATAL(plugin_name, "Could not create socket: %s", strerror(errno));
    }

#if TWS_SOCKTYPE == 1 // for a TCP socket you need to have a three way handshake first
    if (UNLIKELY(connect(sfd, (struct sockaddr*)&server, sizeof(server)) < 0)) {
        T2_PFATAL(plugin_name, "Could not connect to socket: %s, check whether the server side is listening at %s on port %d", strerror(errno), TWS_SERVADD, TWS_DPORT);
    }
#endif // TWS_SOCKTYPE == 1

    tmrk = actTime; // and set our internal send buffer timer
}


/*
 * This callback is only required for sink plugins
 * Refer to parse_binary2text() in utils/bin2txt.c for an example
 */
void t2BufferToSink(outputBuffer_t *buf UNUSED, binary_value_t *bv UNUSED) {
    // check core timer
    if (actTime.tv_sec - tmrk.tv_sec < TWS_TMINTVL || gwzP->wzi == 0) return; // timer expired and is something to print?

    tmrk = actTime; // reset timer

    // Data preparation part

    char srcIP[INET_ADDRSTRLEN];

    char time[MAX_TM_BUF];
    strftime(time, MAX_TM_BUF, "%a %d %b %Y %X", localtime(&tmrk.tv_sec)); // make a nice time string

    // shovel output of tcpWin into sBuf

    int sBufLen = snprintf(sBuf, TWS_MAXSBUF, "# %s\n# IP\tpktTcpCnt\twinRelThCnt\n", time); // send header

    for (int i = 0; i < gwzP->wzi; i++) {
        T2_IPV4_TO_STR(gwzP->wzip[i], srcIP, sizeof(srcIP));       // convert IP to string
        sBufLen += snprintf(&sBuf[sBufLen], TWS_MAXSBUF-sBufLen,      // push into sBuf
                            "%s\t%" PRIu32 "\t%f\n",
                            srcIP, gwzP->tcpCnt[i], gwzP->wzCnt[i]);
    }

    sBuf[sBufLen++] = '\n'; // final separation
    sBuf[sBufLen++] = '\0'; // just make sure that string is always properly terminated

    // Send sBuf via socket

    int32_t written = 0, act_written; // buffer positional indexes
    while (written < sBufLen) { // make sure that the socket sends the whole buffer regardless of flow control socket back pressure
#if TWS_SOCKTYPE == 1
        act_written = write(sfd, sBuf + written, sBufLen - written);
#else // TWS_SOCKTYPE == 0
        act_written = sendto(sfd, sBuf + written, sBufLen - written, 0, (struct sockaddr*)&server, sizeof(server));
#endif // TWS_SOCKTYPE
        if (UNLIKELY(act_written <= 0)) {
            T2_PERR(plugin_name, "Could not send message to socket: %s", strerror(errno));
            if (sfd) close(sfd);
            exit(EXIT_FAILURE);
        }

        written += act_written;
    }
}


/*
 * This function is called once all the packets have been processed.
 * Cleanup all used memory here.
 */
void t2Finalize() {
    if (sfd) close(sfd);
}
